/* Copyright (C) 1998 Oregon Graduate Institute of Science & Technology
 * Copyright (C) 1994, MPEG Software Simulation Group. All Rights Reserved.
 * 
 * See the file COPYRIGHT, which should have been supplied with this
 * package, for licensing details.  We may be contacted through email
 * at <quasar-help@cse.ogi.edu>.
 */

/* idctref.c, Inverse Discrete Fourier Transform, double precision          */

/*  Perform IEEE 1180 reference (64-bit floating point, separable 8x1
 *  direct matrix multiply) Inverse Discrete Cosine Transform
 */


/* Here we use math.h to generate constants.  Compiler results may
   vary a little */

#include "global.h"

#include <math.h>
#include "decoders.h"

#define RB "rb"
#define WB "wb"

#ifndef O_BINARY
#define O_BINARY 0
#endif

#ifndef PI
# ifdef M_PI
#  define PI M_PI
# else
#  define PI 3.14159265358979323846
# endif
#endif

/* private data */

/* cosine transform matrix for 8x1 IDCT */
static double c[8][8];

/* initialize DCT coefficient matrix */

void init_float_idct()
{
  int freq, time;
  double scale;
  DBUG_ENTER();

  for (freq=0; freq < 8; freq++)
  {
    scale = (freq == 0) ? sqrt(0.125) : 0.5;
    for (time=0; time<8; time++)
      c[freq][time] = scale*cos((PI/8.0)*freq*(time + 0.5));
  }
  DBUG_VOID_RETURN();
}

/* perform IDCT matrix multiply for 8x8 coefficient block */

void float_idct(block)
short *block;
{
  int i, j, k, v;
  double partial_product;
  double tmp[64];
  DBUG_ENTER();

  for (i=0; i<8; i++)
    for (j=0; j<8; j++)
    {
      partial_product = 0.0;

      for (k=0; k<8; k++)
        partial_product+= c[k][j]*block[8*i+k];

      tmp[8*i+j] = partial_product;
    }

  /* Transpose operation is integrated into address mapping by switching 
     loop order of i and j */

  for (j=0; j<8; j++)
    for (i=0; i<8; i++)
    {
      partial_product = 0.0;

      for (k=0; k<8; k++)
        partial_product+= c[k][i]*tmp[8*k+j];

      v = floor(partial_product+0.5);
      block[8*i+j] = (v<-256) ? -256 : ((v>255) ? 255 : v);
    }
  DBUG_VOID_RETURN();
}
