/* Copyright (C) 1998 Oregon Graduate Institute of Science & Technology
 * Portions copyright (c) 1995 Brown University.
 * Copyright (c) 1995 The Regents of the University of California.
 * 
 * See the file COPYRIGHT, which should have been supplied with this
 * package, for licensing details.  We may be contacted through email
 * at <quasar-help@cse.ogi.edu>.
 */

/* 
 * gdith.c --
 *
 *      Procedures dealing with grey-scale and mono dithering, 
 *      as well as X Windows set up procedures.
 *
 */

#include "global.h"

#include <math.h>
#include <sys/time.h>

#ifdef __STDC__
#include <stdlib.h>
#include <string.h>
#endif

#include "gdith.h"
#include "video.h"
#include "dither.h"
#include <player/render.h>


/* Range values for lum, cr, cb. */
int LUM_RANGE;
int CR_RANGE;
int CB_RANGE;

/* Array that remaps color numbers to actual pixel values used by X server. */

unsigned char pixel[256];
unsigned long wpixel[256];

/* Arrays holding quantized value ranged for lum, cr, and cb. */

int *lum_values;
int *cr_values;
int *cb_values;

/* Luminance and chrominance lookup tables */
static double *L_tab, *Cr_r_tab, *Cr_g_tab, *Cb_g_tab, *Cb_b_tab;

static void InitColor(void);
static void InitColormap(GdkVisual *visual);
static gboolean try_alloc_colors(GdkColormap *colormap, int ncolors, 
				 GdkColor *colors, gboolean *success);


/*
 *--------------------------------------------------------------
 *
 * InitColor --
 *
 *	Initialize lum, cr, and cb quantized range value arrays.
 *      Also initializes the lookup tables for the possible
 *      values of lum, cr, and cb.
 *    Color values from ITU-R BT.470-2 System B, G and SMPTE 170M
 *    see InitColorDither in 16bits.c for more
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

static void
InitColor(void)
{
  int i, CR, CB;
  DBUG_ENTER();

  L_tab    = (double *)g_malloc(LUM_RANGE*sizeof(double)); 
  Cr_r_tab = (double *)g_malloc(CR_RANGE*sizeof(double));
  Cr_g_tab = (double *)g_malloc(CR_RANGE*sizeof(double));
  Cb_g_tab = (double *)g_malloc(CB_RANGE*sizeof(double));
  Cb_b_tab = (double *)g_malloc(CB_RANGE*sizeof(double));
  for (i=0; i<LUM_RANGE; i++) {
    lum_values[i]  = ((i * 256) / (LUM_RANGE)) + (256/(LUM_RANGE*2));
    L_tab[i] = lum_values[i];
    if (gammaCorrectFlag) {
      L_tab[i] = GAMMA_CORRECTION(L_tab[i]);
    }
  }
  
  for (i=0; i<CR_RANGE; i++) {
    register double tmp;
    if (chromaCorrectFlag) {
      tmp = ((i * 256) / (CR_RANGE)) + (256/(CR_RANGE*2));
      Cr_r_tab[i] = (int) (0.419/0.299) * CHROMA_CORRECTION128D(tmp - 128.0);
      Cr_g_tab[i] = (int) -(0.299/0.419) * CHROMA_CORRECTION128D(tmp - 128.0);
      cr_values[i] = CHROMA_CORRECTION256(tmp);
    } else {
      tmp = ((i * 256) / (CR_RANGE)) + (256/(CR_RANGE*2));
      Cr_r_tab[i] = (int)  (0.419/0.299) * (tmp - 128.0);
      Cr_g_tab[i] = (int) -(0.299/0.419) * (tmp - 128.0);
      cr_values[i] = (int) tmp;
    }
  }

  for (i=0; i<CB_RANGE; i++) {
    register double tmp;
    if (chromaCorrectFlag) {
      tmp = ((i * 256) / (CB_RANGE)) + (256/(CB_RANGE*2));
      Cb_g_tab[i] = (int) -(0.114/0.331) * CHROMA_CORRECTION128D(tmp - 128.0);
      Cb_b_tab[i] = (int)  (0.587/0.331) * CHROMA_CORRECTION128D(tmp - 128.0);
      cb_values[i] = CHROMA_CORRECTION256(tmp);
    } else {
      tmp = ((i * 256) / (CB_RANGE)) + (256/(CB_RANGE*2));
      Cb_g_tab[i] = (int) -(0.114/0.331) * (tmp - 128.0);
      Cb_b_tab[i] = (int)  (0.587/0.331) * (tmp - 128.0);
      cb_values[i] = (int) tmp;
    }
  }

  DBUG_VOID_RETURN();
}


/*
 *--------------------------------------------------------------
 *
 * ConvertColor --
 *
 *	Given a l, cr, cb tuple, converts it to r,g,b.
 *
 * Results:
 *	r,g,b values returned in pointers passed as parameters.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

static void
ConvertColor(l, cr, cb, r, g, b)
     unsigned int l, cr, cb;
     unsigned char *r, *g, *b;
{
  double fl, fcr, fcb, fr, fg, fb;
  DBUG_ENTER();

/*
 * Old method w/o lookup table
 *
 * fl = 1.164*(((double) l)-16.0);
 * fcr =  ((double) cr) - 128.0;
 * fcb =  ((double) cb) - 128.0;
 *
 * fr = fl + (1.366 * fcr);
 * fg = fl - (0.700 * fcr) - (0.334 * fcb);
 * fb = fl + (1.732 * fcb);
 */

  fl = L_tab[l];

  fr = fl + Cr_r_tab[cr];
  fg = fl + Cr_g_tab[cr] + Cb_g_tab[cb];
  fb = fl + Cb_b_tab[cb];

  if (fr < 0.0) fr = 0.0;
  else if (fr > 255.0) fr = 255.0;

  if (fg < 0.0) fg = 0.0;
  else if (fg > 255.0) fg = 255.0;

  if (fb < 0.0) fb = 0.0;
  else if (fb > 255.0) fb = 255.0;

  *r = (unsigned char) fr;
  *g = (unsigned char) fg;
  *b = (unsigned char) fb;

  DBUG_VOID_RETURN();
}

/*
 *--------------------------------------------------------------
 *
 * InitDithering --
 *
 *	Initializes the setup for the right kind of dithering.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

int
InitDithering(void)
{
  int ditherType, best_depth;
  GdkVisual *visual;
  DBUG_ENTER();

  best_depth = gdk_visual_get_best_depth();
  if((best_depth > 8) && 
     (visual = gdk_visual_get_best_with_type(GDK_VISUAL_TRUE_COLOR))) {
    ditherType = FULL_COLOR_DITHER;
  } else if(visual = gdk_visual_get_best_with_type(GDK_VISUAL_PSEUDO_COLOR)) {
    ditherType = FS2_DITHER;  // this is arbirary right now
    InitColor();
  } else {
    // Assume for now that we only support color displays
    DBUG_RETURN(-1);
  } // else
  gtk_widget_push_visual(visual);  // this will ensure that we use this visual when we make the window

  switch (ditherType) {
    
  case HYBRID_DITHER:
    InitHybridDither();
    InitColormap(visual);
    break;
    
  case HYBRID2_DITHER:
    InitHybridErrorDither();
    InitColormap(visual);
    break;
    
  case FS4_DITHER:
    InitFS4Dither();
    InitColormap(visual);
    break;
    
  case FS2_DITHER:
    InitFS2Dither();
    InitColormap(visual);
    break;
    
  case FS2FAST_DITHER:
    InitFS2FastDither();
    break;
    
  case Twox2_DITHER:
    Init2x2Dither();
    InitColormap(visual);
    PostInit2x2Dither();
    break;

#ifdef FIXME
  case GRAY_DITHER:
  case GRAY2_DITHER:
    InitGrayDisplay(dpyname, xinfo);
    break;

  case GRAY256_DITHER:
  case GRAY2562_DITHER:
    InitGray256Display(dpyname, xinfo);
    break;
#endif // FIXME

  case FULL_COLOR_DITHER:
  case FULL_COLOR2_DITHER:
    InitColorDither(visual);
    break;

  case PPM_DITHER:
    InitColorDither(NULL);
    break;

  case ORDERED_DITHER:
    InitOrderedDither();
    InitColormap(visual);
    break;

#ifdef FIXME
  case MONO_DITHER:
  case MONO_THRESHOLD:
    InitMonoDisplay(dpyname, xinfo);
    break;
#endif // FIXME

  case ORDERED2_DITHER:
    InitColormap(visual);
    InitOrdered2Dither();
    break;

  case MBORDERED_DITHER:
    InitColormap(visual);
    InitMBOrderedDither();
    break;
  }
  zz_init_tables();
  DBUG_RETURN(ditherType);
}


static gboolean
try_alloc_colors(GdkColormap *colormap, int ncolors, GdkColor *colors, gboolean *success)
{
  int i;
  gboolean result = FALSE;
  DBUG_ENTER();

  i=gdk_colormap_alloc_colors(colormap, colors, ncolors, FALSE, TRUE, success);
  if(i < ncolors) {
    if(i > 0) {
      gdk_colormap_free_colors(colormap, colors, i);
    } // if
  } else {
    result = TRUE;
  } // else
  DBUG_RETURN(result);
} // try_alloc_colors

static void
InitColormap(GdkVisual *visual)
{
  int ncolors = LUM_RANGE*CB_RANGE*CR_RANGE;
  GdkColor colors[LUM_RANGE*CB_RANGE*CR_RANGE];
  gboolean success[LUM_RANGE*CB_RANGE*CR_RANGE];
  int i, j, lum_num, cr_num, cb_num;
  unsigned char r, g, b;
  GdkColormap *colormap;
  DBUG_ENTER();

  for (i=0; i<ncolors; i++) {
    lum_num = (i / (CR_RANGE*CB_RANGE))%LUM_RANGE;
    cr_num = (i / CB_RANGE)%CR_RANGE;
    cb_num = i % CB_RANGE;
    ConvertColor(lum_num, cr_num, cb_num, &r, &g, &b);
    colors[i].red   = r * 256;
    colors[i].green = g * 256;
    colors[i].blue =  g * 256;
  } // for 
  colormap = gdk_colormap_get_system();
  if((gdk_colormap_get_visual(colormap) == visual) && 
     try_alloc_colors(colormap, ncolors, colors, success)) goto found_cmap;
  colormap = gdk_colormap_new(visual,FALSE);
  if(try_alloc_colors(colormap, ncolors, colors, success)) goto found_cmap;
  colormap = gdk_colormap_new(visual,TRUE);
  if(try_alloc_colors(colormap, ncolors, colors, success)) goto found_cmap;
  fprintf(stderr,"Couldn't allocate colors in colormap!\n");
  exit(-1);
 found_cmap:
  gtk_widget_push_colormap(colormap);
  for (i=0; i<ncolors; i++) {
    pixel[i] = colors[i].pixel;
    wpixel[i] = colors[i].pixel;
  } // for
  DBUG_VOID_RETURN();
} // InitColormap

#if 0

/*
 *--------------------------------------------------------------
 *
 * InitGrayDisplay --
 *
 *	Initialized display for gray scale dither.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitGrayDisplay(name,xinfo)
char *name;
XInfo *xinfo;
{
  int ncolors = 128;
  XColor xcolor;
  int i;
  Colormap dcmap;
  unsigned long tmp_pixels[256];
  Window window;
  Display *display;
  DBUG_ENTER();

  if (!xinfo->doDisplay) DBUG_VOID_RETURN();
  MakeWindow(name,xinfo);
  display=xinfo->display;
  window=xinfo->window;

  xinfo->gc = XCreateGC(display, window, 0, 0);

  dcmap = xinfo->cmap = XDefaultColormap(display, DefaultScreen(display));

  xcolor.flags = DoRed | DoGreen | DoBlue;

  if (xinfo->owncmFlag) goto create_map;
  retry_alloc_grays:
  for (i=0; i<ncolors; i++) {

    xcolor.red = xcolor.green = xcolor.blue = GAMMA_CORRECTION(i*2) * 256;

    if(XAllocColor(display, xinfo->cmap, &xcolor) == 0 
		   && xinfo->cmap == dcmap) {
      int j;
      XWindowAttributes xwa;

      /* Free colors. */
      for(j = 0; j < i; j ++) {
        unsigned long tmp_pixel;
        tmp_pixel = tmp_pixels[j*2];
        XFreeColors(display, xinfo->cmap, &tmp_pixel, 1, 0);
      }

      create_map:
      XGetWindowAttributes(display, window, &xwa);
      xinfo->cmap = XCreateColormap(display, window, xwa.visual, AllocNone);
      XSetWindowColormap(display, window, xinfo->cmap);

      goto retry_alloc_grays;
    }
    tmp_pixels[i*2] = pixel[i*2] = xcolor.pixel;
    tmp_pixels[(i*2)+1] = pixel[(i*2)+1] = xcolor.pixel;
    wpixel[(i*2)] = xcolor.pixel;
    wpixel[(i*2)+1] = xcolor.pixel;
    if(xinfo->depth == 8) {
      wpixel[i*2] |= wpixel[i*2] << 8;
      wpixel[i*2+1] |= wpixel[i*2+1] << 8;
    }
    if(xinfo->depth == 8 || xinfo->depth == 16) {
      wpixel[i*2] |= wpixel[i*2] << 16;
      wpixel[i*2+1] |= wpixel[i*2+1] << 16;
    }
#ifdef SIXTYFOUR_BIT
    if(xinfo->depth == 8 || xinfo->depth == 16 || xinfo->depth == 24 ||
       xinfo->depth == 32) {
      wpixel[i*2] |= wpixel[i*2] << 32;
      wpixel[i*2+1] |= wpixel[i*2+1] << 32;
    }
#endif

  }

  xinfo->ximage = NULL;
  DBUG_VOID_RETURN();
}


/*
 *--------------------------------------------------------------
 *
 * InitGray256Display --
 *
 *	Initialized display for gray scale dither with 256 levels
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */


void InitGray256Display(name, xinfo)
char *name;
XInfo *xinfo;
{
  int ncolors = 256;
  XColor xcolor;
  int i;
  Colormap dcmap;
  int result;
  XWindowAttributes xwa;
  unsigned long tmp_pixels[256];
  Display *display;
  DBUG_ENTER();

  if (!xinfo->doDisplay) DBUG_VOID_RETURN();
  MakeWindow(name,xinfo);
  display=xinfo->display;

  xinfo->gc = XCreateGC(display, xinfo->window, 0, 0);

  dcmap = xinfo->cmap = XDefaultColormap(display, DefaultScreen(display));

  xcolor.flags = DoRed | DoGreen | DoBlue;

  if (xinfo->owncmFlag) {
    XGetWindowAttributes(display, xinfo->window, &xwa);
    xinfo->cmap=XCreateColormap(display, xinfo->window, xwa.visual, AllocNone);
    XSetWindowColormap(display, xinfo->window, xinfo->cmap);
  }

  retry_alloc_grays:
  for (i = 0;  i < ncolors;  i++) {
    xcolor.red = xcolor.green = xcolor.blue = GAMMA_CORRECTION(i) * 256;
    if ((result = XAllocColor(display, xinfo->cmap, &xcolor)) == 0
	 && xinfo->cmap == dcmap) {
      int j;
      unsigned long tmp_pixel;

      /* Free colors. */
      for (j = 0; j < i; j ++) {
        tmp_pixel = tmp_pixels[j];
        XFreeColors(display, xinfo->cmap, &tmp_pixel, 1, 0);
      }

      XGetWindowAttributes(display, xinfo->window, &xwa);
      xinfo->cmap = XCreateColormap(display, xinfo->window, xwa.visual,
				    AllocNone);
      XSetWindowColormap(display, xinfo->window, xinfo->cmap);

      goto retry_alloc_grays;
    }
    tmp_pixels[i] = pixel[i] = xcolor.pixel;
    wpixel[i] = xcolor.pixel;
    if(xinfo->depth == 8)
      wpixel[i] |= wpixel[i] << 8;
    if(xinfo->depth == 8 || xinfo->depth == 16) {
      wpixel[i] |= wpixel[i] << 16;
    }
#ifdef SIXTYFOUR_BIT
    if(xinfo->depth == 8 || xinfo->depth == 16 || xinfo->depth == 24 
       || xinfo->depth == 32) {
      wpixel[i] |= wpixel[i] << 32;
    }
#endif

  }

  xinfo->ximage = NULL;
  DBUG_VOID_RETURN();
}


/*
 *--------------------------------------------------------------
 *
 * InitMonoDisplay --
 *
 *	Initialized display for monochrome dither.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void InitMonoDisplay(name, xinfo)
char *name;
XInfo *xinfo;
{
  XGCValues xgcv;
  Display *display;
  DBUG_ENTER();

  if (!xinfo->doDisplay) DBUG_VOID_RETURN();
  MakeWindow(name,xinfo);
  display=xinfo->display;

  xgcv.background = BlackPixel(display, DefaultScreen(display));
  xgcv.foreground = WhitePixel(display, DefaultScreen(display));

  xinfo->gc = XCreateGC(display, xinfo->window,
			GCForeground | GCBackground,
			&xgcv);

  xinfo->ximage = NULL;
  DBUG_VOID_RETURN();
}


/*
 *--------------------------------------------------------------
 *
 * InitColorDisplay --
 *
 *	Initialized display for full color output.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void
InitColorDisplay(name, xinfo)
char *name;
XInfo *xinfo;
{
  XWindowAttributes winattr;
  DBUG_ENTER();

  if (xinfo == NULL || !xinfo->doDisplay) DBUG_VOID_RETURN();
  if ((xinfo->ditherType == NO_DITHER) || (xinfo->ditherType == PPM_DITHER))
     DBUG_VOID_RETURN();

  MakeWindow(name,xinfo);

  if (xinfo->gc==0) {
    xinfo->gc = XCreateGC(xinfo->display, xinfo->window, 0, 0);
  }
  xinfo->ximage = NULL;

  XGetWindowAttributes(xinfo->display, xinfo->window, &winattr);
  /*
   * Misuse of wpixel 
   */
  wpixel[0] = winattr.visual->red_mask;
  wpixel[1] = winattr.visual->green_mask;
  wpixel[2] = winattr.visual->blue_mask;
  DBUG_VOID_RETURN();
}

#endif

/*
 *--------------------------------------------------------------
 *
 * DoDitherImage --
 *
 *      Called when image needs to be dithered. Selects correct
 *      dither routine based on info in xinfo.ditherType.
 *
 * Results:
 *        The fully decoded and dithered image.
 *
 * Side effects:
 *        None.
 *
 *--------------------------------------------------------------
 */

PictImage *
DoDitherImage(VidDecoder *vid_decoder, GtkWidget *render)
{
 unsigned char *l=vid_decoder->current->luminance,
               *Cr=vid_decoder->current->Cr,
               *Cb=vid_decoder->current->Cb,
               *disp=vid_decoder->current->image->mem;
 int h=(int) vid_decoder->sequence.mb_height * 16;
 int w=(int) vid_decoder->sequence.mb_width * 16;
 int ditherType=vid_decoder->ditherType;
 GdkVisual *visual;
 int matched_depth;
  DBUG_ENTER();

  visual = gtk_widget_get_visual(render);
  matched_depth = visual->depth;
  switch(ditherType) {
  case HYBRID_DITHER:
    HybridDitherImage(l, Cr, Cb, disp, h, w);
    break;

  case HYBRID2_DITHER:
    HybridErrorDitherImage(l, Cr, Cb, disp, h, w);
    break;

  case FS2FAST_DITHER:
    FS2FastDitherImage(l, Cr, Cb, disp, h, w);
    break;

  case FS2_DITHER:
    FS2DitherImage(l, Cr, Cb, disp, h, w);
    break;

  case FS4_DITHER:
    FS4DitherImage(l, Cr, Cb, disp, h, w);
    break;

  case Twox2_DITHER:
    Twox2DitherImage(l, Cr, Cb, disp, h, w);
    break;

  case FULL_COLOR2_DITHER:
    if (matched_depth == 32)
      Twox2Color32DitherImage(l, Cr, Cb, disp, h, w);
    else
      Twox2Color16DitherImage(l, Cr, Cb, disp, h, w);
    break;
  case FULL_COLOR_DITHER:
    if (matched_depth >= 24)
      Color32DitherImage(l, Cr, Cb, disp, h, w);
    else
      Color16DitherImage(l, Cr, Cb, disp, h, w);
    break;

  case GRAY_DITHER:
  case GRAY256_DITHER:
    if (matched_depth == 8) 
      GrayDitherImage(l, Cr, Cb, disp, h, w);
    else if (matched_depth == 16) 
      Gray16DitherImage(l, Cr, Cb, disp, h, w);
    else if (matched_depth == 32 || matched_depth == 24)
      Gray32DitherImage(l, Cr, Cb, disp, h, w);
    break;

  case GRAY2_DITHER:
  case GRAY2562_DITHER:
    if (matched_depth == 8) 
      Gray2DitherImage(l, Cr, Cb, disp, h, w);
    else if (matched_depth == 16) 
      Gray216DitherImage(l, Cr, Cb, disp, h, w);
    else if (matched_depth == 32 || matched_depth == 24)
      Gray232DitherImage(l, Cr, Cb, disp, h, w);
    break;

  case NO_DITHER:
    break;

  case PPM_DITHER:
    Color32DitherImage(l, Cr, Cb, disp, h, w);
    break;

  case ORDERED_DITHER:
    OrderedDitherImage(l, Cr, Cb, disp, h, w);
    break;

  case MONO_DITHER:
    MonoDitherImage(l, Cr, Cb, disp, h, w);
    break;

  case MONO_THRESHOLD:
    MonoThresholdImage(l, Cr, Cb, disp, h, w);
    break;

  case ORDERED2_DITHER:
    Ordered2DitherImage(l, Cr, Cb, disp, h, w);
    break;

  case MBORDERED_DITHER:
    MBOrderedDitherImage(l, Cr, Cb, disp, h, w, vid_decoder->ditherFlags);
    break;
  }

  DBUG_RETURN(ReturnPictImage(vid_decoder));
}


#if 0
extern char *strrchr();
#define PPM_BITS 8


/*
 *--------------------------------------------------------------
 *
 * ExecutePPM --
 *
 *	Write out a display plane as a PPM file.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */

void
ExecutePPM(img, vid_decoder)
  PictImage *img;
  VidDecoder *vid_decoder;
{
  static int munged = 0;
  static char mungedInputName[300];
  char fileName[300];
  FILE *file;
  int n;
  unsigned int *p;
  unsigned int r, g, b;
  DBUG_ENTER();

  if (!munged) {
    char *cp;

    cp = strrchr(vid_decoder->filename, '/');
    if (cp != NULL)
      ++cp;
    else
      cp = vid_decoder->filename;
    strcpy(mungedInputName, cp);
    cp = strrchr(mungedInputName, '.');
    if (cp != NULL)
	*cp = '\0';
    munged = 1;
  }

  sprintf(fileName, "%s_%05d.ppm", mungedInputName, vid_decoder->totNumFrames);
  file = fopen(fileName, "w");
  if (file == NULL) {
    perror(fileName);
    exit(1);
  }

  fprintf(file, "P6\n");
  fprintf(file, "%d %d\n", vid_decoder->sequence.h_size,
      vid_decoder->sequence.v_size);
  fprintf(file, "255\n");

  p = (unsigned int *) img->display;
  n = vid_decoder->sequence.h_size * vid_decoder->sequence.v_size;
  while (n > 0) {
    r = *p & 0xff;
    g = (*p >> PPM_BITS) & 0xff;
    b = (*p >> (2*PPM_BITS)) & 0xff;
    putc(r, file);
    putc(g, file);
    putc(b, file);
    ++p;
    --n;
  }

  fclose(file);
  DBUG_VOID_RETURN();
}
#endif
