/* Copyright (C) 1998 Oregon Graduate Institute of Science & Technology
 * Portions copyright (c) 1995 Brown University.
 * Copyright (c) 1995 The Regents of the University of California.
 * 
 * See the file COPYRIGHT, which should have been supplied with this
 * package, for licensing details.  We may be contacted through email
 * at <quasar-help@cse.ogi.edu>.
 */

/* 
 * stream_play.c --
 *
 *      This file contains C code that implements a straight-forward
 *      sequential decoder.
 *
 */

#include <syslog.h>

#include "global.h"
#include "stream_play.h"
#include "video.h"
#include "readfile.h"
#include "gdith.h"
#include "util.h"
#include <player/render.h>
#ifdef HAVE_LAUNCHER
#include <pipeline/init_modules.h>
#include <getopt.h>
#endif /* HAVE_LAUNCHER */



typedef void * (*pthread_start_routine) (void *);

/* Declarations of functions. */
#ifdef HAVE_LIBPTHREAD
static void *DisplayLoop(VidStream *str);
#ifdef HAVE_LIBSWQ
static void AddVidStreamExit(VidStream *);
static void RemoveVidStreamExit(VidStream *);
#endif
#endif
static void DoPictureDisplay(VidStream *);
static void SendPictureToDisplay(VidStream *, PictImage *, int);
static void FlushPictureDisplay(VidStream *);
static int ComputeFrameDelay(VidStream *, int);
static void DoDelay(VidStream *, int);

/* Video rates table */
/* Cheat on Vid rates, round to 30, and use 30 if illegal value 
   Except for 9, where Xing means 15, and given their popularity, we'll
   be nice and do it */
static int VidRateNum[16]={30, 24, 24, 25, 30, 30, 50, 60, 
                         60, 15, 30, 30, 30, 30, 30, 30};



/*
 *--------------------------------------------------------------
 *
 * InitPictImageRing --
 *
 *	Sets up the ring structure for pict images.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
void
InitPictImageRing(ring, cnt)
  PictImageRing *ring;
  int cnt;
{
  DBUG_ENTER();
  ring->frame = NULL;
  ring->used = NULL;
  ring->frame_delay = NULL;
  ring->size = cnt;
  DBUG_VOID_RETURN();
}




/*
 *--------------------------------------------------------------
 *
 * FillPictImageRing --
 *
 *	Fills the PictImageRing with PictImages, now that the
 *	correct information is known to initialize them.
 *
 * Results:
 *      Success code.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
BOOLEAN
FillPictImageRing(PictImageRing *ring, VidDecoder *vid_decoder, GtkWidget *render)
{
  PictImage *frame = ring->frame;
  int *used = ring->used;
  int size = ring->size;
  int i;
  DBUG_ENTER();

  if (frame) {
    for (i = 0; i < size; ++i)
      DestructPictImage(&frame[i]);
  }
  else {
    ring->frame = frame = g_malloc(size * sizeof(PictImage));
    ring->used = used = g_malloc(size * sizeof(int));
    ring->frame_delay = g_malloc(size * sizeof(int));
  }

  /* XXX Move this to before Destruct?  (And in PictImageRing Destruct?) */

  for (i = 0; i < size; ++i) {
    InitPictImage(&frame[i], vid_decoder, render);
  }

  /* Clear bitmap */
  memzero(used, size * sizeof(int));

  DBUG_RETURN(TRUE);
}



/*
 *--------------------------------------------------------------
 *
 * DestructPictImageRing --
 *
 *	Destroys the PictImages stored in the ring.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
void
DestructPictImageRing(PictImageRing *ring)
{
  PictImage *frame = ring->frame;
  int size = ring->size;
  int i;
  DBUG_ENTER();

  if (frame == NULL)
    DBUG_VOID_RETURN();

  /* Destroy allocated items */

  for (i = 0; i < size; i++) {
    DestructPictImage(&frame[i]);
  }

  free(frame);
  free(ring->frame_delay);
  free(ring->used);

  frame = NULL;
  DBUG_VOID_RETURN();
}



/*
 *--------------------------------------------------------------
 *
 * AllocPictImage --
 *
 *	Grabs a free PictImage from the ring.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
PictImage *
AllocPictImage(ring)
  PictImageRing *ring;
{
  PictImage *frame = ring->frame;
  int *used = ring->used;
  int size = ring->size;
  int i;
  DBUG_ENTER();

  if (!frame)
    DBUG_RETURN(NULL);

  for (;;) {
    for (i = 0; i < size; ++i) {
      if (!used[i]) {
	used[i] = 1;
	DBUG_RETURN(&frame[i]);
      }
    }
    usleep(5000);
  }
}



/*
 *--------------------------------------------------------------
 *
 * UsePictImage --
 *
 *	Increases the ref count on a PictImage.
 *
 *--------------------------------------------------------------
 */
void
UsePictImage(PictImageRing *ring,  PictImage *image)
{
  int pos = image - ring->frame;
  DBUG_ENTER();
  assert(pos >= 0 && pos < ring->size);

  ++ring->used[pos];
  DBUG_VOID_RETURN();
}




/*
 *--------------------------------------------------------------
 *
 * ReleasePictImage --
 *
 *	Returns a used PictImage to the ring of free picts
 *
 * Results:
 *      None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
void
ReleasePictImage(PictImageRing *ring, PictImage *image)
{
  int pos = image - ring->frame;
  DBUG_ENTER();
  assert(pos >= 0 && pos < ring->size);
  assert(ring->used[pos]);

  --ring->used[pos];
  DBUG_VOID_RETURN();
}




/*
 *--------------------------------------------------------------
 *
 * GetPictImageFrameDelay --
 *
 *	Gets the frame delay value associated with a picture.
 *
 *--------------------------------------------------------------
 */
int
GetPictImageFrameDelay(PictImageRing *ring, PictImage *image)
{
  int pos = image - ring->frame;
  DBUG_ENTER();
  assert(pos >= 0 && pos < ring->size);

  DBUG_RETURN(ring->frame_delay[pos]);
}




/*
 *--------------------------------------------------------------
 *
 * SetPictImageFrameDelay --
 *
 *	Sets the frame delay value associated with a picture.
 *
 *--------------------------------------------------------------
 */
void SetPictImageFrameDelay(PictImageRing *ring, PictImage *image, int delay)
{
  int pos = image - ring->frame;
  DBUG_ENTER();
  assert(pos >= 0 && pos < ring->size);

  ring->frame_delay[pos] = delay;
  DBUG_VOID_RETURN();
}



/*
 *--------------------------------------------------------------
 *
 * NewVidStream --
 *
 *	Creates and initializes a VidStream structure.
 *
 * Results:
 *	The newly created VidStream.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

VidStream *
NewVidStream(DataStream *data_stream, char *inputName, int ditherType)
{
  VidStream *new;
  DBUG_ENTER();

  new = (VidStream *) g_malloc(sizeof(VidStream));
  /* Initialize subfields */

  new->stream = data_stream;
  new->ditherType = ditherType;
  InitVidDecoder(&new->decoder, inputName, ditherType);
  InitPictImageRing(&new->ring, RING_BUF_SIZE);

  /* Initialize pointers to image spaces. */

  new->past = new->future = NULL;

  /* Initialize other fields */
  new->film_has_started = FALSE;
  new->film_has_ended = FALSE;
  new->rate_ticks = 0;
  new->frame_rate = 1;

#ifdef HAVE_LIBPTHREAD
#ifdef HAVE_LIBSWQ
  new->pic_queue_in = new->pic_queue_out = 0;

  AddVidStreamExit(new);
  new->queue_region.sr_addr.sa_tag = swq_sysvshmem;
  new->queue_region.sr_addr.u.sysv.sa_key = 0;
    
  new->pic_queue_in = swqCreate(&new->queue_region,
	QUEUE_SIZE, SWQ_PRODUCER | 0600);
  if (!new->pic_queue_in) {
    perror("swqCreate");
    exit(1);
  }
#endif /* HAVE_LIBSWQ */
  if (pthread_create(&new->display_thread, NULL, (pthread_start_routine)DisplayLoop, (void *) new)) {
    perror("pthread_create");
    exit(1);
  }
#endif /* HAVE_LIBPTHREAD */
  DBUG_RETURN(new);
}



/*
 *--------------------------------------------------------------
 *
 * DestroyVidStream --
 *
 *	Destroys a VidStream structure.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *	None.
 *
 *--------------------------------------------------------------
 */
void
DestroyVidStream(VidStream *astream)
{
#if HAVE_LIBPTHREAD && HAVE_LIBSWQ
  swqRegion null_data;
#endif
#if HAVE_CHANNELS
  dbuf *dbuf;
#endif
  DBUG_ENTER();

#ifdef HAVE_LIBPTHREAD
#ifdef HAVE_LIBSWQ
  /* Kill subthread: */
  /*   First, tell it to die... */

  null_data.sr_size = 0;
  null_data.sr_addr.sa_tag = swq_absolute;
  null_data.sr_addr.u.abs.sa_ptr = NULL;
  if (swqSend(astream->pic_queue_in, &null_data, 0) < 0) {
    perror("swqSend");
    exit(1);
  }
#endif /* HAVE_LIBSWQ */

  /*   ...then clean up the corpse. */

  if (pthread_join(astream->display_thread, NULL)) {
    perror("pthread_join\n");
    exit(1);
  }

#ifdef HAVE_LIBSWQ
  /* Close shared queue. */

  if(swqClose(astream->pic_queue_in) < 0) {
    astream->pic_queue_in = 0;
    perror("swqClose(in)");
    exit(1);
  }
  astream->pic_queue_in = 0;

  RemoveVidStreamExit(astream);
#endif
#endif /* HAVE_LIBPTHREAD */

#ifdef HAVE_CHANNELS
  dbuf = new_raw_dbuf(sizeof(void *));
  dbGROW(dbuf,sizeof(void *));
  *(void **)dbCUR(dbuf) = NULL;
  chSEND_free(astream->stream->channelout,dbuf);
#endif

  /* Destruct subfields */

  DestructVidDecoder(&astream->decoder);
  DestructPictImageRing(&astream->ring);

  /* Free memory. */

  free(astream);
  DBUG_VOID_RETURN();
}



/*
 *--------------------------------------------------------------
 *
 * ResetVidStream --
 *
 *	Re-initializes a VidStream structure. Takes
 *      as parameter a pointer to the VidStream to reset.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *      None.
 *
 *--------------------------------------------------------------
 */

void
ResetVidStream(VidStream *vid)
{
  PictImage *pic;
  DBUG_ENTER();

  /* Reset our state */

  SeekStream(vid->stream, 0);
  ResetVidDecoder(&vid->decoder);

  pic = ReturnPictImage(&vid->decoder);
  if (pic)
    ReleasePictImage(&vid->ring, pic);

  /* Initialize pointers to image spaces. */

  if (vid->past) {
    ReleasePictImage(&vid->ring, vid->past);
    vid->past = NULL;
  }
  if (vid->future) {
    ReleasePictImage(&vid->ring, vid->future);
    vid->future = NULL;
  }

  /* We are at the beginning of the film, so film has not ended... */
  vid->film_has_ended = FALSE;
  vid->film_has_started = FALSE;  /* Or started. */
  DBUG_VOID_RETURN();
}



/*
 *--------------------------------------------------------------
 *
 * mpegVidRsrc --
 *
 *      Parses bit stream until some concrete element
 *	(frame, gop header, start/end header.  Do everything
 *	from decoding to dithering to display.
 *
 * Results:
 *      A value indicating whether the movie is OK.
 *
 * Side effects:
 *      Bit stream is irreversibly parsed. If a picture is completed,
 *      a function is called to display the frame at the correct time.
 *
 *--------------------------------------------------------------
 */

BOOLEAN
mpegVidRsrc(VidStream *vid_stream)
{
  DataStream *data_stream;
  VidDecoder *vid_decoder;
  PictImage *pic;
  static GtkWidget *window;
  unsigned int data;
  int status, factor;
  DBUG_ENTER();

  if (vid_stream == NULL)
    DBUG_RETURN(FALSE);

  data_stream = vid_stream->stream;
  vid_decoder = &vid_stream->decoder;

  /*
   * If called for the first time, find start code, make sure it is a
   * sequence start code.
   */

  if (!vid_stream->film_has_started) {
    next_start_code(data_stream);  /* sets curBits */
    show_bits32(data);
    if (data != SEQ_START_CODE) {
      fprintf(stderr, "This is not an MPEG video stream. (%x)\n",data);
      DBUG_RETURN(FALSE);
    }
  }

  /* XXX If we had data in different location than header, reposition here. */

  for (;;) {
    /* Get next 32 bits (size of start codes). */

    show_bits32(data);

    /*
     * Process according to start code (or parse macroblock if not a
     * start code at all).
     */

    switch (data) {

    case SEQ_END_CODE:
    case ISO_11172_END_CODE:
    case 0:

      /* Display future frame, if any. */

      FlushPictureDisplay(vid_stream);

      /* Sequence done. Do the right thing. */

      vid_stream->film_has_ended=TRUE;
      goto done;
      break;

    case SEQ_START_CODE:

      if (vid_stream->film_has_started) {
	/* XXX Should incorporate new picture information here. */

	flush_bits(24);
	goto skip_picture;
      }

      /* Sequence start code. Parse sequence header. */

      if (ParseSequence(data_stream, &vid_decoder->sequence) != PARSE_OK)
	goto error;

      StartVidSeq(vid_decoder);

      /*
       * Initialize ring buffer of pict images now that dimensions
       * of image space are known.
       */

      factor = 1 + IS_2x2_DITHER(vid_stream->ditherType);
      vid_stream->render = 
	gui_render_new(vid_decoder->filename,
		       vid_decoder->sequence.h_size * factor, 
		       vid_decoder->sequence.v_size * factor);
      gui_mainwin_add_render(vid_stream->render);

      if (!FillPictImageRing(&vid_stream->ring, vid_decoder, vid_stream->render))
	goto error;

      vid_stream->film_has_started = TRUE;

      /* Compute frame rate. */
      vid_stream->frame_rate = VidRateNum[vid_decoder->sequence.picture_rate];

      if (vid_stream->rate_ticks < 0) {
	gettimeofday(&vid_stream->tflast, (struct timezone *)NULL);
	vid_stream->rate_ticks = 1000000 / vid_stream->frame_rate;
	vid_stream->pts = 0;
      }

      /*
       * Return after sequence start code so that application above can use
       * info in header.
       */

      goto done;

    case GOP_START_CODE:

      /* Group of Pictures start code. Parse group header. */
      {
	int hours = vid_decoder->group.tc_hours;
	int minutes = vid_decoder->group.tc_minutes;
	int seconds = vid_decoder->group.tc_seconds;
	int pictures = vid_decoder->group.tc_pictures;

	if (vid_decoder->group.drop_flag) {
	  if (++pictures >= vid_stream->frame_rate) {
	    pictures -= vid_stream->frame_rate;
	    if (++seconds >= 60) {
	      seconds -= 60;
	      if (++minutes >= 60) {
		minutes -= 60;
		++hours;
	      }
	    }
	  }
	}

	if (ParseGoP(data_stream, &vid_decoder->group) != PARSE_OK)
	  goto error;

	hours -= vid_decoder->group.tc_hours;
	if (hours > 0)
	  goto renew_tc;
	if (hours)
	  minutes += hours * 60;
	
	minutes -= vid_decoder->group.tc_minutes;
	if (minutes > 0)
	  goto renew_tc;
	if (minutes)
	  seconds += minutes * 60;

	seconds -= vid_decoder->group.tc_seconds;
	if (seconds > 0)
	  goto renew_tc;
	if (seconds)
	  pictures += seconds * vid_stream->frame_rate;

	pictures -= vid_decoder->group.tc_pictures;
	if (pictures >= 0)
	  goto renew_tc;

	vid_stream->temp_ref += pictures;
	vid_stream->fut_temp_ref += pictures;	/* Yet-to-be-displayed frame */
      }
      break;

renew_tc:
      vid_stream->fut_temp_ref -= vid_stream->temp_ref + 1;
      vid_stream->temp_ref = -1;
      break;

    case PICTURE_START_CODE:

      /* Picture start code. Parse picture header and first slice header. */

      status = ParsePicture(data_stream, &vid_decoder->picture);

      if (status == PARSE_OK) {
	/* Find a pict image structure in ring buffer not currently locked. */

	PictImage *current = vid_decoder->current ? vid_decoder->current :
	  AllocPictImage(&vid_stream->ring);
	assert(current);

	if (vid_decoder->picture.code_type == P_TYPE) {
	  status = StartPicture(vid_decoder, current, vid_stream->future, NULL);
	}
	else {
	  status = StartPicture(vid_decoder, current, vid_stream->past,
	      vid_stream->future);
	}
      }
	  
      if (status == SKIP_PICTURE)
	goto skip_picture;
      else if (status != PARSE_OK)
	goto error;

      show_bits32(data);
      if (data < SLICE_MIN_START_CODE || data >= SLICE_MAX_START_CODE)
	goto error;

      break;

    case SEQUENCE_ERROR_CODE:
      flush_bits32;
      next_start_code(data_stream);
      goto done;
      
    default:

      /* Check for slice start code. */

      if ((data >= SLICE_MIN_START_CODE) && (data <= SLICE_MAX_START_CODE)) {

	if (!vid_decoder->current) {	/* Make sure we are in a picture. */
	  flush_bits32;
	  goto error;
	}

	/* Slice start code. Parse slice header. */

	if (ParseSlice(data_stream, &vid_decoder->slice) != PARSE_OK)
	  goto error;

	StartSlice(vid_decoder);
      }
      goto slice;
    }
    continue;

  slice:
    /* Parse macroblocks. */

    for (;;) {
      /* Check to see if actually a startcode and not a macroblock. */

      if (next_bits(23, 0x00000000, data_stream))
	goto end_mb;

	/* Not start code. Parse Macroblock. */

	if (ParseDecodeMB(data_stream, vid_decoder) != PARSE_OK)
	  goto error;
    }

  end_mb:
    /* Not macroblock, actually start code. Get start code. */

    next_start_code(data_stream);
    show_bits32(data);

    /*
     * If start code is outside range of slice start codes, frame is
     * complete, display frame.
     */

    if (((data < SLICE_MIN_START_CODE) || (data > SLICE_MAX_START_CODE)) &&
	(data != SEQUENCE_ERROR_CODE)) {

      // hack
      DoPictureDisplay(vid_stream);
      goto done;
    }
    continue;
  }

error:
  fprintf(stderr, "Error!!!!\n");	/* XXX Make noise optional. */
  next_start_code(data_stream);
  DBUG_RETURN(FALSE);

skip_picture:
  /* Clear picture being decoded. */
  pic = ReturnPictImage(vid_decoder);
  if (pic)
    ReleasePictImage(&vid_stream->ring, pic);

  /* Find next major start code. */
  for (;;) {
    next_start_code(data_stream);
    show_bits32(data);
    if ((data == PICTURE_START_CODE) || (data == GOP_START_CODE) ||
	(data == SEQ_END_CODE) || (data == ISO_11172_END_CODE))
      break;
    flush_bits(24);
  }
  goto done;

done:

  DBUG_RETURN(TRUE);

}



/*
 *--------------------------------------------------------------
 *
 * DoPictureDisplay --
 *
 *	Converts image from Lum, Cr, Cb to colormap space. Puts
 *      image in lum plane. Updates past and future frame
 *      pointers. Dithers image. Sends to display mechanism.
 *
 * Results:
 *	Pict image structure locked if displaying or if frame
 *      is needed as past or future reference.
 *
 * Side effects:
 *	Lum plane pummelled.
 *
 *--------------------------------------------------------------
 */

static void
DoPictureDisplay(VidStream *vid_stream)
{
  PictImage *current;
  DBUG_ENTER();

  /* Convert to colormap space and dither. */

  current = DoDitherImage(&vid_stream->decoder, vid_stream->render);

  /* Update past and future references if needed. */

  if ((vid_stream->decoder.picture.code_type == I_TYPE) ||
      (vid_stream->decoder.picture.code_type == P_TYPE)) {
    if (vid_stream->future == NULL) {
      vid_stream->future = current;
      vid_stream->fut_temp_ref = vid_stream->decoder.picture.temp_ref;
    } else {
      int past_temp_ref = vid_stream->fut_temp_ref;

      if (vid_stream->past)
        ReleasePictImage(&vid_stream->ring, vid_stream->past);
      vid_stream->past = vid_stream->future;
      vid_stream->future = current;
      vid_stream->fut_temp_ref = vid_stream->decoder.picture.temp_ref;

      UsePictImage(&vid_stream->ring, vid_stream->past);
      SendPictureToDisplay(vid_stream, vid_stream->past, past_temp_ref);
    }
  } else {
    SendPictureToDisplay(vid_stream, current, 
			 vid_stream->decoder.picture.temp_ref);
  }
  DBUG_VOID_RETURN();
}



/*
 *--------------------------------------------------------------
 *
 * FlushPictureDisplay --
 *
 *	Sends last picture to display mechanism, if any.
 *
 *--------------------------------------------------------------
 */

static void
FlushPictureDisplay(VidStream *vid_stream)
{
  PictImage *last = vid_stream->future;
  DBUG_ENTER();

  if (vid_stream->future) {
    int last_temp_ref = vid_stream->fut_temp_ref;

    if (vid_stream->past)
      ReleasePictImage(&vid_stream->ring, vid_stream->past);
    vid_stream->past = NULL;
    vid_stream->future = NULL;

    SendPictureToDisplay(vid_stream, last, last_temp_ref);
  }
  DBUG_VOID_RETURN();
}



/*
 *--------------------------------------------------------------
 *
 * SendPictureToDisplay --
 *
 *	Sends a picture to be displayed.  The picture is ref count
 *	decremented afterwards.
 *
 *--------------------------------------------------------------
 */

static void
SendPictureToDisplay(VidStream *vid_stream,  PictImage *current,  int temp_ref)
{
  int frame_delay;
  DBUG_ENTER();

  frame_delay = ComputeFrameDelay(vid_stream, temp_ref);
  DBUG_PRINT("display","Send frame with delay %d\n", frame_delay);
  vid_stream->pts += vid_stream->rate_ticks / 1000 * frame_delay;

#if defined(HAVE_LIBSWQ) || defined(HAVE_CHANNELS)
  SetPictImageFrameDelay(&vid_stream->ring, current, frame_delay);

#ifdef HAVE_LIBSWQ
  {
    swqRegion region;

    region.sr_size = 1;
    region.sr_addr.sa_tag = swq_absolute;
    region.sr_addr.u.abs.sa_ptr = (char *) current;
    if (swqSend(vid_stream->pic_queue_in, &region, 0) < 0) {
      perror("swqSend");
      exit(1);
    }
  }
#else  // HAVE_CHANNELS
  dbTTS(current->dbuf) = vid_stream->pts;
  chSEND_free(vid_stream->stream->channelout, current->dbuf);
#endif // HAVE_CHANNELS
#else // ! (HAVE_LIBSWQ || HAVE_CHANNELS) 
  DoDelay(vid_stream, frame_delay);
  vid_stream->decoder.totNumFrames++;
  gui_render_display_image(vid_stream->render, current->image);
  ReleasePictImage(&vid_stream->ring, current);
#endif // ! (HAVE_LIBSWQ || HAVE_CHANNELS) 

  DBUG_VOID_RETURN();
}


/*
 *--------------------------------------------------------------
 *
 * ComputeFrameDelay --
 *
 *	Computes the apropriate amount of time to wait to make
 *	video play at the correct speed.
 *
 * Results:
 *	The number of frames to wait.
 *
 *--------------------------------------------------------------
 */

static int
ComputeFrameDelay(vid_stream, temp_ref)
  VidStream *vid_stream;
  int temp_ref;
{
  int frames;
  DBUG_ENTER();

  if (vid_stream->rate_ticks <= 0)
    DBUG_RETURN(0);

  frames = temp_ref - vid_stream->temp_ref;
  vid_stream->temp_ref = temp_ref;

  DBUG_RETURN(frames);
}




/*
 *--------------------------------------------------------------
 *
 * DoDelay --
 *
 *	Waits an apropriate amount of time to make video play
 *	at the correct speed.
 *
 * Results:
 *	Waits for an amount of time based upon rate_ticks.
 *
 *--------------------------------------------------------------
 */

static void
DoDelay(vid_stream, frames)
  VidStream *vid_stream;
  int frames;
{
  struct timeval tf, tfnow;
  int usec, sec;
  DBUG_ENTER();

  if (frames <= 0)
    DBUG_VOID_RETURN();

  /* Setup target for this frame */
  /* TODO: no no no.  this is too jittery!  
     If a frame is late,

     Target should be next_tick + (frames - 1) * rate_ticks,
     so that targets are always exactly start_time + (x * rate_ticks) for
     some integer x */
  tf.tv_usec = vid_stream->tflast.tv_usec + vid_stream->rate_ticks * frames;
  tf.tv_sec = vid_stream->tflast.tv_sec;
  while (tf.tv_usec >= 1000000) {
    tf.tv_usec -= 1000000;
    ++tf.tv_sec;
  }

  gettimeofday(&tfnow, (struct timezone *)NULL);
  usec = tf.tv_usec - tfnow.tv_usec;
  sec  = tf.tv_sec - tfnow.tv_sec;
  if (usec < 0) {
    usec += 1000000;
    sec--;
  }
  
  /* If we're not behind, wait a bit */
  /* also, if we're close to the target, say X% of a rate_tick, then
     go ahead an display anyway */
  if ((sec >= 0)  &&  usec > 0) {
    DBUG_PRINT("display","picture on time\n");
    tfnow.tv_sec = sec;
    tfnow.tv_usec = usec;
    select(0, NULL, NULL, NULL, &tfnow); 
    vid_stream->tflast = tf;
  }
  else {
    DBUG_PRINT("display","late picture\n");
    vid_stream->tflast = tf;
    vid_stream->tflast = tfnow;
  }
  DBUG_VOID_RETURN();
}





static void
usage()
{
        fprintf(stderr, 
                "Usage: mpegdisplay [--best-effort]  if mpegdisplay appears after mpegdecode \"\\|\"\n"
		);
        exit(1);
}


#ifdef HAVE_CHANNELS

#ifdef HAVE_LAUNCHER
void
mpegdisplay(int argc, char **argv) {
  
  channel_src *channelin = NULL;
  
  VidStream *stream;
  PictImage *current;
  int best_effort = FALSE;
  dbuf *dbuf;

  DBUG_ENTER();
  


 argc = module_init(argc, argv, &channelin, ARG_chunk, NULL, ARG_null);

 // getopt_long doesn't support multi-threading, thus we need to use a mutex.
  pthread_mutex_lock(&launcher_mutex);

  while (1) {
    int option_index = 0;
    struct option long_options[] =
    {
      {"help", 0, 0, 0},
      {"best-effort", 0, 0, 0},
      {0, 0, 0, 0}
    };
    int c = getopt_long(argc, argv, "",long_options, &option_index);
    if (c == -1)
      break;
    switch (c) {
    case 0:
      switch(option_index) {
      case 0:
	usage();
	break;
      case 1:
	best_effort = TRUE;
	break;
      } // swictch
      break;
    default:
      usage();
    }
  }

  optind = 0;  // to "reset" getopt!!!
 // we are done with getopt so we can release the mutex.
 pthread_mutex_unlock(&launcher_mutex);

 if (channelin == NULL)
   usage();

  
#else

void
mpegdisplay(channel_src *channelin)
{
  VidStream *stream;
  PictImage *current;
  dbuf *dbuf;
  DBUG_ENTER();

#endif /* HAVE_LAUNCHER */

  DBUG_TRACE_FNS("ExecuteDisplay",NULL);
  DBUG_KEYWORDS("display",NULL);
  DBUG_SET_FLAGS(DCTL_DBG | DCTL_TIM | DCTL_TRA | DCTL_PID);
  dbuf = chRECV(channelin);

  stream = *(VidStream **)dbCUR(dbuf);
  dbFREE(dbuf);
  if(!stream) DBUG_VOID_RETURN();

  for (;;) {
    DBUG_PRINT("display","chRECV\n");
    dbuf = chRECV(channelin);
    if((!dbuf) || (!dbLEN(dbuf))) break;
    current = *(PictImage **)dbCUR(dbuf);
    if (!current) {
      break;
    }
    if(!best_effort) {
      DoDelay(stream, GetPictImageFrameDelay(&stream->ring, current));
      stream->decoder.totNumFrames++;
      gui_render_display_image(stream->render, current->image);
    } else {
      // For benchmarking, cost of display isn't that interesting (?)
      // works better remotely...
      stream->decoder.totNumFrames++;
    } // if
    ReleasePictImage(&stream->ring, current);
  } // for
  stream->film_has_ended = FALSE; // signal decoder that its OK to leave
  gui_mainwin_remove();
  DBUG_VOID_RETURN();
}
#endif // HAVE_CHANNELS

#ifdef HAVE_LIBPTHREAD
/*
 *--------------------------------------------------------------
 *
 * DisplayLoop --
 *
 *	Loop to display incoming frames at the right rate.
 *
 *--------------------------------------------------------------
 */

static void *
DisplayLoop(VidStream *stream)
{
  DBUG_ENTER();
  DBUG_TRACE_FNS("ExecuteDisplay",NULL);
  DBUG_KEYWORDS("display",NULL);
  DBUG_SET_FLAGS(DCTL_DBG | DCTL_TIM | DCTL_TRA | DCTL_PID);

#ifdef HAVE_LIBSWQ
  stream->pic_queue_out = swqOpen(&stream->queue_region, SWQ_CONSUMER | 0600);
  if (!stream->pic_queue_out) {
    perror("swqOpen(out)");
    exit(1);
  }

  for (;;) {
    PictImage *current;
    swqRegion region;

    DBUG_PRINT("display","swqRecv\n");
    if (swqRecv(stream->pic_queue_out, &region, 0) < 0) {
      perror("swqRecv");
      exit(1);
    }

    assert(region.sr_addr.sa_tag == swq_absolute);
    current = (PictImage *) region.sr_addr.u.abs.sa_ptr;
    if (!current)
      break;

    DoDelay(stream, GetPictImageFrameDelay(&stream->ring, current));
    stream->decoder.totNumFrames++;
    gui_render_display_image(stream->render, current->image);
    ReleasePictImage(&stream->ring, current);
  }

  if (swqClose(stream->pic_queue_out) < 0) {
    stream->pic_queue_out = 0;
    perror("swqClose(out)");
    exit(1);
  }
  stream->pic_queue_out = 0;
#else /* ! HAVE_LIBSWQ */
#error Using pthreads requires the use of libswq.
#endif /* HAVE_LIBSWQ */

  DBUG_RETURN(NULL);
}


#ifdef HAVE_LIBSWQ

/*
 *--------------------------------------------------------------
 *
 * VidStreamExit --
 *
 *	Clean up vid streams upon exit.
 *
 *--------------------------------------------------------------
 */

static VidStream **cleanup_stream = 0;
static int cleanup_stream_count;

static void
VidStreamExit(void)
{
  int i;
  DBUG_ENTER();

  if (!cleanup_stream)
    DBUG_VOID_RETURN();

  for (i = 0; i < cleanup_stream_count; ++i) {

    if (!cleanup_stream[i])
      continue;

    if (swqClose(cleanup_stream[i]->pic_queue_in) < 0) {
      cleanup_stream[i]->pic_queue_in = 0;
      perror("swqClose(in)");
      exit(1);
    }
    cleanup_stream[i]->pic_queue_in = 0;
    
    if (swqClose(cleanup_stream[i]->pic_queue_out) < 0) {
      cleanup_stream[i]->pic_queue_out = 0;
      perror("swqClose(out)");
      exit(1);
    }
    cleanup_stream[i]->pic_queue_out = 0;
  }
  DBUG_VOID_RETURN();
}



/*
 *--------------------------------------------------------------
 *
 * AddVidStreamExit --
 *
 *	Mark the supplied vid stream to be cleaned up upon exit.
 *
 *--------------------------------------------------------------
 */

static void
AddVidStreamExit(stream)
  VidStream *stream;
{
  VidStream **tmp;
  int i;
  DBUG_ENTER();

  /* Create the storage space, if needed. */

  if (!cleanup_stream) {
    cleanup_stream_count = 4;
    cleanup_stream = calloc(cleanup_stream_count, sizeof(VidStream *));

    if (!cleanup_stream) {
      perror("calloc");
      exit(1);
    }

    atexit(VidStreamExit);
  }

  /* Find a slot. */

  for (i = 0; i < cleanup_stream_count; ++i) {
    if (cleanup_stream[i])
      continue;

    cleanup_stream[i] = stream;
    DBUG_VOID_RETURN();
  }

  /* No space!  Expand. */

  tmp = realloc(cleanup_stream,
      (cleanup_stream_count << 1) * sizeof(VidStream *));
  if (!tmp) {
    perror("realloc");
    exit(1);
  }
  cleanup_stream_count <<= 1;

  /* Save value now... */

  tmp[i] = stream;
  ++i;

  /* ...and clear rest of the space. */

  for (; i < cleanup_stream_count; ++i)
    cleanup_stream[i] = 0;
  DBUG_VOID_RETURN();
}



/*
 *--------------------------------------------------------------
 *
 * RemoveVidStreamExit --
 *
 *	Remove the supplied vid stream from those to be cleaned
 *	up upon exit.
 *
 *--------------------------------------------------------------
 */

static void
RemoveVidStreamExit(stream)
  VidStream *stream;
{
  int i;
  DBUG_ENTER();

  assert(!cleanup_stream);

  /* Find it's slot. */

  for (i = 0; i < cleanup_stream_count; ++i) {
    if (cleanup_stream[i] != stream)
      continue;

    cleanup_stream[i] = 0;
    DBUG_VOID_RETURN();
  }

  /* Not there! Error! */

  fprintf(stderr, "RemoveVidStreamExit: unknown stream!\n");
  abort();
  DBUG_VOID_RETURN();
}
#endif /* HAVE_LIBSWQ */
#endif /* HAVE_LIBPTHREAD */

